`timescale 1ns / 100ps

/*
All data from the computer comes in through the USB_IO module, where it is then decoded and sent to the opcode processor.
Data to the computer comes from the USB_IO module, but it comes from both the opcode processor and from the data capture block.
All data that is not from the data retrieval process comes from the opcode processor.
*/

module USB_IO(	IF_CLK,
					FPGA_RESET,

			    	SLAVE_READ,
			    	SLAVE_WRITE,
			    	SLAVE_OUTPUT_ENABLE,
		   	 	FIFO_EMPTY_FLAG_C,
             	FIFO_FULL_FLAG_B,
				 	FIFO_ADDRESS,
					PACKET_END,
			    	
			    	DATA_FROM_USB_PORT,
			    	DATA_FROM_USB_TO_OPCODE_PROCESSOR,
		   	 	DATA_FROM_DDC,
				 	DATA_TO_USB_PORT,

				 	USB_READ_DATA_AVAILABLE,
		       	READ_DATA_PROCESSED,

		       	DDC_DATA_AVAILABLE_FOR_WRITE,
		       	USB_WRITE_DATA_REQUESTED_FROM_DDC_FIFO,
		       	DDC_DATA_LOADED_FOR_WRITE,
		    		USB_WRITE_DATA_PROCESSED,
		    		PKTEND_INHIBITOR,
		    		PKTEND_SENT_ACK,
		    		SEND_PKTEND

					/*
					OPCODE_DATA_AVAILABLE_FOR_WRITE,
		       	USB_WRITE_OPCODE_DATA_REQUESTED_FROM_DDC_FIFO,
		       	OPCODE_DATA_LOADED_FOR_WRITE,
		    		USB_WRITE_OPCODE_DATA_PROCESSED
					*/
		    		);
input IF_CLK;
input FPGA_RESET;
output SLAVE_READ;
reg    SLAVE_READ;
output SLAVE_WRITE;
reg    SLAVE_WRITE;
output SLAVE_OUTPUT_ENABLE;
reg    SLAVE_OUTPUT_ENABLE;
input FIFO_EMPTY_FLAG_C;
input FIFO_FULL_FLAG_B;


input [15:0] DATA_FROM_USB_PORT;

output [15:0] DATA_FROM_USB_TO_OPCODE_PROCESSOR;
reg [15:0] DATA_FROM_USB_TO_OPCODE_PROCESSOR;

input [15:0] DATA_FROM_DDC;

output [15:0] DATA_TO_USB_PORT;
reg [15:0] DATA_TO_USB_PORT;	   

output [1:0] FIFO_ADDRESS;
reg [1:0] FIFO_ADDRESS;

output USB_READ_DATA_AVAILABLE;
reg USB_READ_DATA_AVAILABLE;

input READ_DATA_PROCESSED;
reg read_data_processed_gated;

input DDC_DATA_AVAILABLE_FOR_WRITE;
//reg ddc_data_available_for_write_gated;

output USB_WRITE_DATA_REQUESTED_FROM_DDC_FIFO;
reg USB_WRITE_DATA_REQUESTED_FROM_DDC_FIFO;

input DDC_DATA_LOADED_FOR_WRITE;
reg ddc_data_loaded_for_write_gated;

output USB_WRITE_DATA_PROCESSED;
reg USB_WRITE_DATA_PROCESSED;

output PACKET_END;
reg PACKET_END;

input PKTEND_INHIBITOR;
reg pktend_inhibitor_gated;

output PKTEND_SENT_ACK;
reg PKTEND_SENT_ACK;

input SEND_PKTEND;
reg send_pktend_gated;


reg [4:0] FSM, FSMn;
parameter IDLE = 5'h00,
          IDLE_FIFO_ADDRESS_INCR = 5'h01,

		BUFFER_NOT_EMPTY = 5'h0A,
		READ_DATA = 5'h0B,
		HOLD_UNTIL_FPGA_READY = 5'h0C,
		ADVANCE_FIFO = 5'h0D,
		CLEANUP = 5'h0E,

		WAIT_FOR_FIFO_FULL_FLAG_B_TO_GO_HIGH = 5'h1A,
      WAIT_FOR_PKTEND_SENT_ACK = 5'h1B,

		REQUEST_DATA_TO_BE_WRITTEN = 5'h11,
		PULSE_SLWR = 5'h12,
    	WORD_WRITTEN_NOW_WAIT = 5'h13,
    	EVALUATE_WRITE_STATUS = 5'h14,
		COMMIT_PACKET = 5'h15;

reg slave_output_enable_i;
reg slave_read_i;
reg slave_write_i;
reg read_parallel_data_trigger_i;
reg fifo_address_incr_i;
reg load_data_onto_port_i;
reg USB_read_data_available_i;
reg USB_write_data_requested_i;
reg USB_write_data_processed_i;
reg packet_end_i;
reg pktend_sent_ack_i;

always @(negedge IF_CLK or negedge FPGA_RESET)   // data on fifo bus updates on rising edge
  begin
  if(FPGA_RESET==1'b0)
    begin
    FSM=IDLE;
    SLAVE_OUTPUT_ENABLE=1'b1;
    SLAVE_READ=1'b1;
    SLAVE_WRITE=1'b1;
	 FIFO_ADDRESS[1:0]=2'b00;
    DATA_FROM_USB_TO_OPCODE_PROCESSOR[15:0]=16'b0000000000000000;
	 DATA_TO_USB_PORT[15:0]=16'b0000000000000000;
    USB_READ_DATA_AVAILABLE=1'b0;
    USB_WRITE_DATA_REQUESTED_FROM_DDC_FIFO=1'b0;
    USB_WRITE_DATA_PROCESSED=1'b0;
    PACKET_END=1'b1;
	 ddc_data_loaded_for_write_gated=1'b0;
	 read_data_processed_gated=1'b0;
	 //ddc_data_available_for_write_gated=1'b0;
	 pktend_inhibitor_gated=1'b0;
	 PKTEND_SENT_ACK=1'b0;
	 send_pktend_gated=1'b0;
    end
  else
    begin
    FSM=FSMn;
    SLAVE_OUTPUT_ENABLE=slave_output_enable_i;
    SLAVE_READ=slave_read_i;
    SLAVE_WRITE=slave_write_i;
    if(read_parallel_data_trigger_i==1'b1) DATA_FROM_USB_TO_OPCODE_PROCESSOR[15:0]=DATA_FROM_USB_PORT[15:0];
    //if(fifo_address_incr_i==1'b1) FIFO_ADDRESS[0]=~FIFO_ADDRESS[0];	  //uncomment to access fifos 4 and 8
    //if(fifo_address_incr_i==1'b1 && FIFO_ADDRESS[0]==1'b1) FIFO_ADDRESS[1]=~FIFO_ADDRESS[1];	//uncomment to access fifos 4 and 8
	 if(fifo_address_incr_i==1'b1) FIFO_ADDRESS[1]=~FIFO_ADDRESS[1]; //comment out to access fifos 4 and 8
	 FIFO_ADDRESS[0]=FIFO_ADDRESS[0];
    if(load_data_onto_port_i==1'b1) DATA_TO_USB_PORT[15:0]=DATA_FROM_DDC[15:0];
    USB_READ_DATA_AVAILABLE=USB_read_data_available_i;
    USB_WRITE_DATA_REQUESTED_FROM_DDC_FIFO=USB_write_data_requested_i;
    USB_WRITE_DATA_PROCESSED=USB_write_data_processed_i;
    PACKET_END=packet_end_i;
	 ddc_data_loaded_for_write_gated=DDC_DATA_LOADED_FOR_WRITE;
	 read_data_processed_gated=READ_DATA_PROCESSED;
	 //ddc_data_available_for_write_gated=DDC_DATA_AVAILABLE_FOR_WRITE;
	 pktend_inhibitor_gated=PKTEND_INHIBITOR;
	 PKTEND_SENT_ACK=pktend_sent_ack_i;
	 send_pktend_gated=SEND_PKTEND;
    end
  end

always @(	FSM or 
		FIFO_EMPTY_FLAG_C or
		FIFO_ADDRESS or
		read_data_processed_gated or
		FIFO_FULL_FLAG_B or
		//ddc_data_available_for_write_gated or
		DDC_DATA_AVAILABLE_FOR_WRITE or
		ddc_data_loaded_for_write_gated or
		pktend_inhibitor_gated or
		send_pktend_gated
		)
  begin
  FSMn=FSM;
  slave_read_i=1'b1;
  slave_write_i=1'b1;
  slave_output_enable_i=1'b1;
  read_parallel_data_trigger_i=1'b0;
  fifo_address_incr_i=1'b0;
  load_data_onto_port_i=1'b0;
  USB_read_data_available_i=1'b0;
  USB_write_data_requested_i=1'b0;
  USB_write_data_processed_i=1'b0;
  packet_end_i=1'b1;
  pktend_sent_ack_i=1'b0;

  case(FSM)

    IDLE:  //00
    	begin
	    if(FIFO_EMPTY_FLAG_C==1'b1 && FIFO_ADDRESS[1]==1'b0)	 // if 1 then read buffer is not empty.
	     begin
	     //FSMn=BUFFER_NOT_EMPTY;
	     FSMn = READ_DATA;
	     slave_output_enable_i=1'b0;
	     end
	   else if(DDC_DATA_AVAILABLE_FOR_WRITE==1'b1 && FIFO_ADDRESS[1]==1'b1 && FIFO_FULL_FLAG_B==1'b1) //if fifo full, then we can't write
	     begin
		     FSMn=REQUEST_DATA_TO_BE_WRITTEN;
			  //USB_write_data_requested_i=1'b1;
	     end
	   else if(send_pktend_gated==1'b1 && FIFO_ADDRESS[1]==1'b1)
	     begin
            FSMn=WAIT_FOR_FIFO_FULL_FLAG_B_TO_GO_HIGH;
	     end    
	   else
	     begin
	     FSMn=IDLE_FIFO_ADDRESS_INCR;
	     fifo_address_incr_i=1'b1;
	     end
	   end

    IDLE_FIFO_ADDRESS_INCR:  //01
      begin
	     FSMn=IDLE;
	   end

    //////////////////////////////////
    //  WRITE DATA FROM USB BUFFER  //
    //////////////////////////////////

    REQUEST_DATA_TO_BE_WRITTEN:  //11
      begin
	   USB_write_data_requested_i=1'b1;
	   if(ddc_data_loaded_for_write_gated==1'b1)  //data from fifo is now both available and on port
	     begin
	     load_data_onto_port_i=1'b1;
		  FSMn=PULSE_SLWR;
	     end
	   else  //data is not on fpga fifo output port yet
	     begin
	     FSMn=REQUEST_DATA_TO_BE_WRITTEN;
	     end
	   end

    PULSE_SLWR:  //12
      begin
	   slave_write_i	= 1'b0;
	   FSMn=WORD_WRITTEN_NOW_WAIT;
	   end
    
    WORD_WRITTEN_NOW_WAIT:  //13
      begin
		  if(ddc_data_loaded_for_write_gated==1'b1)
		    begin
			   USB_write_data_processed_i=1'b1;
			   FSMn=WORD_WRITTEN_NOW_WAIT;
			 end
		  else
		    begin
	   		FSMn=EVALUATE_WRITE_STATUS;
			 end
	   end

    EVALUATE_WRITE_STATUS:  //14
      begin
	   if(DDC_DATA_AVAILABLE_FOR_WRITE==1'b1)
	     begin
	     FSMn=IDLE;
	     end
	   else
	     begin
	     FSMn=COMMIT_PACKET;
	     end
	   end

    COMMIT_PACKET:  //15
      begin
	   if(pktend_inhibitor_gated==1'b0) packet_end_i=1'b0;
	   FSMn=IDLE;
	   end





    ////////////////////////////
    //  FORCE PKTEND TO USB   //
    ////////////////////////////

    WAIT_FOR_FIFO_FULL_FLAG_B_TO_GO_HIGH: //1A
    begin
        if(FIFO_FULL_FLAG_B==1'b1)
			  begin
					FSMn=WAIT_FOR_PKTEND_SENT_ACK;
					packet_end_i=1'b0;
			  end
        else
			  begin
					FSMn=WAIT_FOR_FIFO_FULL_FLAG_B_TO_GO_HIGH;
			  end
    end
    
    WAIT_FOR_PKTEND_SENT_ACK:  //1B
    begin
        if(send_pktend_gated==1'b0)
			  begin
					FSMn=IDLE;
			  end
        else
			  begin
					FSMn=WAIT_FOR_PKTEND_SENT_ACK;
					pktend_sent_ack_i=1'b1;
			  end
    end
    

    

    /////////////////////////////////
    //  READ DATA FROM USB BUFFER  //
    /////////////////////////////////

    BUFFER_NOT_EMPTY: //0A
      begin
	   slave_output_enable_i=1'b0;
	   FSMn=READ_DATA;
	   end

    READ_DATA:  //0B
    	begin 	    
	   slave_output_enable_i=1'b0;
	   read_parallel_data_trigger_i=1'b1;
	   FSMn=HOLD_UNTIL_FPGA_READY;
	   end

    HOLD_UNTIL_FPGA_READY:  //0C
      begin
	   slave_output_enable_i=1'b0;
	   USB_read_data_available_i=1'b1;
	   if(read_data_processed_gated==1'b1)
	     begin
	     FSMn=CLEANUP;
	     slave_read_i=1'b0;
	     end
	   else
	     begin
	     FSMn=HOLD_UNTIL_FPGA_READY;
	     end
	   end

    ADVANCE_FIFO:  //0D
    	begin
	   slave_output_enable_i=1'b0;
	   FSMn=CLEANUP;
	   end

    CLEANUP:  //0E
    	begin
	   slave_output_enable_i=1'b0;
	   if(FIFO_EMPTY_FLAG_C==1'b1) // 1 means not empty
	     begin
	     FSMn=READ_DATA;
	     end
	   else
	     begin
	     FSMn=IDLE;
	     end
	   end

    default:
    	begin
	   FSMn=IDLE;
	   end

    endcase

  end
		









/*
always @(negedge interface_clock or posedge FPGA_RESET)
  begin
  if(FPGA_RESET)
    begin
    SLAVE_READ=1'b1;
    SLAVE_WRITE=1'b1;
    SLAVE_OUTPUT_ENABLE=1'b1;
    end
  else
    begin
    SLAVE_READ=1'b1;
    SLAVE_WRITE=1'b1;
    SLAVE_OUTPUT_ENABLE=1'b1;
    if(FIFO_EMPTY_FLAG_C == 1'b1)	    // is fifo empty?  if 1, then it is not empty yet.
      begin
      SLAVE_READ=1'b0;
      SLAVE_OUTPUT_ENABLE=1'b0;
      end
    if(FIFO_FULL_FLAG_B == 1'b1)      // is fifo full?  if 0, then it is full.
      begin
	 SLAVE_WRITE=1'b0;
	 end
    end
  end
*/

endmodule