`timescale 1ns / 100ps
// ===========================================================================
//                    Copyright 2002
//          Texas Instruments Deutschland GmbH
//                 All rights reserved
//
// Filename:          memory.v
// Modelname:         MEMORY
// Title:             Memory controller for IDT71V424
// Purpose:           Synthesis				 
// Comment:            Gate-equivalents
// Assumptions:       x
// Limitations:       x
// Specification ref: ADS7869
// ---------------------------------------------------------------------------
// Modification History:
// ---------------------------------------------------------------------------
// Version | Author |  Date  | Changes
// --------|--------|--------|------------------------------------------------
//  1.0    |  Bad   |24.04.02| Original Code
// --------|--------|--------|------------------------------------------------
//  1.1    |  Bad   |25.04.02| The addresses are managed on the toplevel
// --------|--------|--------|------------------------------------------------
//  1.2    |  Bad   |02.07.02| Add WAIT for calling device
// --------|--------|--------|------------------------------------------------
// ===========================================================================

//`timescale 1 ns / 1 ns

module MEMORY (CLK, RESET, WR_N, RD_N, DIN, DOUT, RAM_DATA,
               RAM_CS_N, RAM_OE_N, RAM_WE_N, RAM_WAIT);

input         CLK;
input         RESET;
input         WR_N;
input         RD_N;
input   [15:0] DIN;
output  [15:0] DOUT;
inout   [15:0] RAM_DATA;
output        RAM_CS_N;
output        RAM_OE_N;		  
output        RAM_WE_N;
output        RAM_WAIT;

wire    [15:0] DOUT;
tri     [15:0] RAM_DATA;
wire          RAM_CS_N;
wire          RAM_OE_N;
wire          RAM_WE_N;
wire          RAM_WAIT;

// ---------------------------------

wire          DATA_OE_N;

reg     [15:0] RAMDATA_OUT_Q;
reg     [15:0] RAMDATA_OUT_D;
//reg     [15:0] DIN_Q;
//reg     [15:0] DIN_D;
reg     [15:0] DOUT_Q;
reg     [15:0] DOUT_D;

reg     [4:0] STATE_Q;
reg     [4:0] STATE_D;

// ---------------------------------

assign RAM_DATA[15:0]     = (DATA_OE_N == 1'b0) ? RAMDATA_OUT_Q[15:0] : 16'bzzzzzzzz_zzzzzzzz;
assign DOUT[15:0]         = DOUT_Q[15:0];
assign RAM_WAIT     = STATE_Q[0];
assign DATA_OE_N    = STATE_Q[1];
assign RAM_WE_N     = STATE_Q[2];
assign RAM_CS_N     = STATE_Q[3];
assign RAM_OE_N     = STATE_Q[4];

// ---------------------------------

//                           RAM_WAIT -------+
//                          DATA_OE_N ------+|
//                           RAM_WE_N -----+||
//                           RAM_CS_N ----+|||
//                           RAM_OE_N ---+||||
parameter WAIT_FOR_START            = 5'b11110;
parameter ACTIVATE_WE_N             = 5'b10001;
parameter DEACTIVATE_CS_N           = 5'b11001;
parameter WAIT_FOR_END              = 5'b11111;
parameter ACTIVATE_CS_N             = 5'b10111;
parameter ACTIVATE_OE_N             = 5'b00111;

always @(STATE_Q or WR_N or RD_N or RAMDATA_OUT_Q or
         DOUT_Q or DIN or RAM_DATA)
begin
  RAMDATA_OUT_D <= RAMDATA_OUT_Q;
  DOUT_D <= DOUT_Q;
  begin
    case (STATE_Q)
      WAIT_FOR_START:
        begin
          if (WR_N == 1'b0)
            begin
              RAMDATA_OUT_D <= DIN;
              STATE_D <= ACTIVATE_WE_N;  
            end  
          else
            if (RD_N == 1'b0)
              STATE_D <= ACTIVATE_CS_N;
            else
              STATE_D <= WAIT_FOR_START;
        end
      ACTIVATE_WE_N:
        begin
          STATE_D <= DEACTIVATE_CS_N;  
        end
      DEACTIVATE_CS_N:
        begin
          STATE_D <= WAIT_FOR_END;  
        end
      WAIT_FOR_END:
        begin
          if ((RD_N == 1'b1) & (WR_N == 1'b1))
            STATE_D <= WAIT_FOR_START;  
          else  
            STATE_D <= WAIT_FOR_END;
        end
      ACTIVATE_CS_N:
        begin
          STATE_D <= ACTIVATE_OE_N;  
        end
      ACTIVATE_OE_N:
        begin
          DOUT_D <= RAM_DATA;
          STATE_D <= WAIT_FOR_END;  
        end
      default:
        begin
          STATE_D <= WAIT_FOR_START;  
        end
    endcase    
  end
end




// ---------------------------------

always @(posedge CLK or negedge RESET)
  begin
    if(RESET == 0)
      begin
        RAMDATA_OUT_Q <= 16'b0;
        STATE_Q <= WAIT_FOR_START;
//        DIN_Q <= 16'b0;
        DOUT_Q <= 16'b0;
      end  
    else
      begin
        RAMDATA_OUT_Q <= RAMDATA_OUT_D;
        STATE_Q <= STATE_D;
//        DIN_Q <= DIN_D;
        DOUT_Q <= DOUT_D;  
      end  
  end

endmodule
